/*
 * Copyright (C) 2002 Stealth.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Stealth.
 * 4. The name Stealth may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <linux/module.h>
#include <errno.h>
#include <sys/ptrace.h>
#include <sys/stat.h>
#include "cap.h"

extern int wait(int *);

int try_chown()
{
	char p[] = "/tmp/fooXXXXXX";
	int r, fd = mkstemp(p);
	if (fd < 0)
		return 0;
	close(fd);

	/* try a give-away */
	if (chown(p, getuid()+1, getgid()+1) < 0)
		r = 0;
	else
		r  = 1;

	unlink(p);
	return r;
}


int try_setuid()
{
	int euid = geteuid();

	if (seteuid(euid + 1) < 0)
		return 0;

	seteuid(euid);
	return 1;
}


int try_setgid()
{
	int egid = getegid();

	if (setegid(egid + 1) < 0)
		return 0;

	setegid(egid);
	return 1;
}


int try_kill()
{
	/* XXX: suffices? */
	if (kill(1, SIGCONT) < 0)
		return 0;
	return 1;
}


int try_bind()
{
	struct sockaddr_in sin;
	int r, fd = socket(PF_INET, SOCK_STREAM, 0);
	if (fd < 0)
		return 0;
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(666);

	if (bind(fd, (struct sockaddr*)&sin, sizeof(sin)) < 0)
		r = 0;
	else
		r = 1;

	close(fd);
	return r;
}

int try_net_raw()
{
	int fd = socket(PF_INET, SOCK_RAW, 0);

	if (fd >= 0) {
		close(fd);
		return 1;
	}
	return 0;
}

	
int try_nice()
{
	return (nice(-1) == 0);
}


extern caddr_t create_module(const char *, size_t);

int try_module()
{
	errno = 0;
	create_module("adore", 1234);
	delete_module("adore");
	return (errno == 0);
}



int try_chroot()
{
	int r;
	if (fork() == 0) {
		if (chroot("/tmp") < 0)
			exit(0);
		else
			exit(1);
	}
	wait(&r);
	return r != 0;
}


int try_rawio()
{
	int fd = open("/dev/kmem", O_RDONLY);
	if (fd < 0)
		return 0;
	close(fd);
	return 1;
}


int try_admin()
{
	char h[1024];
	memset(h, 0, sizeof(h));
	gethostname(h, sizeof(h));
	if (sethostname("hola!", 5) < 0)
		return 0;
	sethostname(h, strlen(h));
	return 1;
}


int try_net_admin()
{
	int sock;
	struct ifreq ifr;

	strcpy(ifr.ifr_name, "lo");

	if ((sock = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
		 return 0;

	if (ioctl(sock, SIOCGIFFLAGS, &ifr) < 0)
		return 0;

	ifr.ifr_flags &= ~IFF_UP;
	if (ioctl(sock, SIOCSIFFLAGS, &ifr) < 0)
		return 0;

	ifr.ifr_flags |= IFF_UP;
	ioctl(sock, SIOCSIFFLAGS, &ifr);
	close(sock);
	return 1;
}


int try_ptrace()
{
	int child, r = 0;

	if ((child = fork()) == 0) {
		sleep(10);
		exit(0);
	}
	if (ptrace(PTRACE_ATTACH, child, 0, 0) < 0)
		r = 0;
	else
		r = 1;
	kill(child, SIGKILL);
	wait(NULL);
	return r;
}


int try_mknod()
{
	unlink("/tmp/fd0");
	if (mknod("/tmp/fd0", 0600|S_IFCHR, 2<<8) < 0)
		return 0;
	unlink("/tmp/fd0");
	return 1;
}



struct {
	int value;
	char *name;
	int (*try)();
} caps[] = {
	{0, "CAP_CHOWN", try_chown},
	{1, "CAP_DAC_OVERRIDE", NULL},
	{2, "CAP_DAC_READ_SEARCH", NULL},
	{3, "CAP_FOWNER", NULL},
	{4, "CAP_FSETID", NULL},
	{5, "CAP_KILL", try_kill},
	{6, "CAP_SETGID", try_setgid},
	{7, "CAP_SETUID", try_setuid},
	{8, "CAP_SETPCAP", NULL},
	{9, "CAP_LINUX_IMMUTABLE", NULL},
	{10, "CAP_NET_BIND_SERVICE", try_bind},
	{11, "CAP_NET_BROADCAST", NULL},
	{12, "CAP_NET_ADMIN", try_net_admin},
	{13, "CAP_NET_RAW", try_net_raw},
	{14, "CAP_IPC_LOCK", NULL},
	{15, "CAP_IPC_OWNER", NULL},
	{16, "CAP_SYS_MODULE", try_module},
	{17, "CAP_SYS_RAWIO", try_rawio},
	{18, "CAP_SYS_CHROOT", try_chroot},
	{19, "CAP_SYS_PTRACE", try_ptrace},
	{20, "CAP_SYS_PACCT", NULL},
	{21, "CAP_SYS_ADMIN", try_admin},
	{22, "CAP_SYS_BOOT", NULL},//haha :>
	{23, "CAP_SYS_NICE", try_nice},
	{24, "CAP_SYS_RESOURCE", NULL},
	{25, "CAP_SYS_TIME", NULL},
	{26, "CAP_SYS_TTY_CONFIG", NULL},
	{27, "CAP_MKNOD", try_mknod},
	{28, "CAP_LEASE", NULL},
	{-1, (void*)0}
};


/* if (capable(d.cap_effective, CAP_SYS_MODULE)
 * ...
 */
int capable(int cap, int flag)
{
	return (cap & (1<<flag));
}


int print_cap(cap_user_data_t new, cap_user_data_t old)
{
	int i = 0;
	FILE *f;
	
	if (!new || !old)
		return -1;

	f = fopen("/dev/tty", "w+");
	if (!f)
		return -1;

	fprintf(f, "nE %x nI %x nP %x\n"
		   "oE %x oI %x oP %x\n\n", 
		new->effective, new->inheritable, new->permitted,
		old->effective, old->inheritable, old->permitted);


	/* Print New's advanced (effective) caps over old ones */
	/* HACK! This is left here due to a private version of capcan */
	for (i = 0; caps[i].value != -1; ++i) {
		if (capable(new->effective, caps[i].value) &&
		    !capable(old->effective, caps[i].value))
			fprintf(f, "e %d %s\n", caps[i].value, caps[i].name);
	}

	printf("\n");

	/* Print New's advanced (inhertiable) caps over old ones */
	for (i = 0; caps[i].value != -1; ++i) {
		if (capable(new->inheritable, caps[i].value) &&
		    !capable(old->inheritable, caps[i].value))
			fprintf(f, "i %d %s\n", caps[i].value, caps[i].name);
	}


	/* No news */
	if (new->effective == new->permitted)
		return 0;

	printf("\n");

	/* Print New's advanced permitted caps */
	for (i = 0; caps[i].value != -1; ++i) {
		if (capable(new->permitted, caps[i].value) &&
		    !capable(old->permitted, caps[i].value))
			fprintf(f, "p %d %s\n", caps[i].value, caps[i].name);
	}

	fclose(f);	
	return 0;
}

int brute_caps()
{
	int i = 0;

	for (; caps[i].value != -1; ++i) {
		if (caps[i].try) {
			if (caps[i].try()) {
				printf("b %d %s\n", caps[i].value,
						    caps[i].name);
			}
		}
	}
	return 0;
}

	
